/*
 * XFireworks 1.0 - pyrotechnics simulation program
 * Copyright (c) 1999-2004 Rony B Chandran <ronybc@asia.com>
 *
 * For latest version and to drop a small donation,
 * visit http://www.ronybc.8k.com
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <X11/Xlib.h>

#define WIDTH 640
#define HEIGHT 340
#define SHELLCOUNT 3
#define PIXCOUNT 500                   /* 500  */ 
#define RNDLIFE 2000                   /* 1200 */
#define MINLIFE 500                    /* 500  */
#define POWER 5                        /* 5    */
#define FTWEAK 12                      /* 12   */

static int light_on = 1;

static int width;
static int height;
static unsigned char *palaka1;
static unsigned char *palaka2;
static Display *display;
static Window win;
static XImage *xim; 

typedef struct {
  unsigned int burn;
  float x;
  float y;
  float xv;
  float yv;}firepix;

typedef struct {
  unsigned int life;
  unsigned int color;
  unsigned int cx,cy;
  unsigned int special;
  unsigned int cshift;
  float air,lum;
  firepix *fpix; }fireshell;

int rnd(int x) {
  return(random()%x);}   /* causes 'divide-by-zero */

int explode(fireshell *fs){
  float air,adg = 0.001;     /* gravity */
  unsigned int n,c;
  unsigned int h = height;
  unsigned int w = width;
  firepix *fp = fs->fpix;
  if(fs->cshift) --fs->cshift;
  if((fs->cshift+1)%50==0) fs->color = ~fs->color;
  c = fs->color;
  air = fs->air;
  fs->lum *= 0.99;
  for(n=PIXCOUNT;n;n--){
  if(fp->burn){ --fp->burn; 
  if(fs->special){
  fp->x += fp->xv = fp->xv * air + (float)(rnd(200)-100)/2000;
  fp->y += fp->yv = fp->yv * air + (float)(rnd(200)-100)/2000 +adg; }
  else{
  fp->x += fp->xv = fp->xv * air + (float)(rnd(200)-100)/20000;
  fp->y += fp->yv = fp->yv * air + adg; }
  if(fp->y > h) fp->yv *= -0.24;
  if(fp->x < w && fp->x > 0 && fp->y < h && fp->y > 0){
  unsigned int *prgb = (unsigned int *)(palaka1 + ((int)fp->y * w + (int)fp->x)*4);
               *prgb = c; }
  } fp++;
  } return(--(fs->life)); }

void recycle(fireshell *fs,int x,int y){
  unsigned int n,pixlife;
  firepix *fp = fs->fpix;
  fs->cx = x;
  fs->cy = y;
  fs->color = (rnd(155)+100) <<16 |
              (rnd(155)+100) <<8  |
               rnd(255);
  fs->life = rnd(RNDLIFE)+MINLIFE;
  fs->air  = 1-(float)(rnd(200))/10000;
  fs->lum  = 20000;                 /* millions of jouls..! */
  fs->cshift  = !rnd(5) ? 120:0; 
  fs->special = !rnd(10) ? 1:0; 
  pixlife = rnd(fs->life)+fs->life/10+1;    /* ! */
  for(n=0;n<PIXCOUNT;n++){
  fp->burn = rnd(pixlife)+32;
  fp->xv = POWER*(float)(rnd(20000)-10000)/10000;
  fp->yv = sqrt(POWER*POWER - fp->xv * fp->xv) *
               (float)(rnd(20000)-10000)/10000;
  fp->x = x;
  fp->y = y; 
  fp++;             }}

void blur(void){
  unsigned int n;
  unsigned int w = width;
  unsigned int h = height;
  unsigned char *pa, *pb, *pm;
  pm = palaka1;
  for(n=w*4;n;n--){ *pm = 0; pm++; }    /* clean first line */
  h-=2; 
  pa = pm-(w*4);
  pb = pm+(w*4);
  for(n=w*h*4;n;n--){
  *pm=(unsigned char) 
  (((int)*(pm+4) + (int)*(pm-4) + (int)*pa + (int)*pb)/4); /* *0.98 */
  pa++; pm++; pb++; }
  for(n=w*4;n;n--){ *pm = 0; pm++; }}   /* clean last line */

void fast_fade(void){
  unsigned int n,mask = 0xfefefefe;
  unsigned int *pl = (unsigned int *) palaka1;
  for(n=width*height;n;n--){
  *pl = (*pl & mask) >>1;
  pl++; }}

void fade(int v){
  unsigned int n;
  float f = 1 - (float)v/255;
  unsigned char *pl = palaka1;
  for(n=width*height*4;n;n--){
  *pl = *pl * f; pl++;}}

void light(fireshell *fss){
  unsigned int l,t,n,x,y;
  float s;
  int w = width;
  int h = height;
  unsigned char *dim = palaka2;
  unsigned char *sim = palaka1;
  fireshell *f;
  for(y=0;y<h;y++){
  for(x=0;x<w;x++){
  f = fss; s = 0;
  for(n=SHELLCOUNT;n;n--,f++){
  s += f->lum / ( 1 + sqrt((f->cx - x)*(f->cx - x)+(f->cy - y)*(f->cy - y))); }
  l = s;
  t = l + *sim;
  *dim = (unsigned char) ( t>255 ? 255 : t);
  t = l + *(sim+1);
  *(dim+1) = (unsigned char) ( t>255 ? 255 : t);
  t = l + *(sim+2);
  *(dim+2) = (unsigned char) ( t>255 ? 255 : t);
  sim += 4; dim += 4; }}}

void light_2x2(fireshell *fss){
  unsigned int l,t,n,x,y;
  float s;
  int w = width;
  int h = height;
  unsigned char *dim = palaka2;
  unsigned char *sim = palaka1;
  int nl = w*4;
  fireshell *f;
  for(y=0;y<h;y+=2){
  for(x=0;x<w;x+=2){
  f = fss; s = 0;
  for(n=SHELLCOUNT;n;n--,f++){
  s += f->lum / ( 1 + sqrt((f->cx - x)*(f->cx - x)+(f->cy - y)*(f->cy - y))); }
  l = s;
  t = l + *sim;
  *dim = (unsigned char) ( t>255 ? 255 : t);        /* cmov's */
  t = l + *(sim+1);
  *(dim+1) = (unsigned char) (t>255 ? 255:t);
  t = l + *(sim+2);
  *(dim+2) = (unsigned char) (t>255 ? 255:t);

  t = l + *(sim+4);
  *(dim+4) = (unsigned char) (t>255 ? 255:t);
  t = l + *(sim+5);
  *(dim+5) = (unsigned char) (t>255 ? 255:t);
  t = l + *(sim+6);
  *(dim+6) = (unsigned char) (t>255 ? 255:t);

  t = l + *(sim+nl);
  *(dim+nl) = (unsigned char) (t>255 ? 255:t);
  t = l + *(sim+nl+1);
  *(dim+nl+1) = (unsigned char) (t>255 ? 255:t);
  t = l + *(sim+nl+2);
  *(dim+nl+2) = (unsigned char) (t>255 ? 255:t);

  t = l + *(sim+nl+4);
  *(dim+nl+4) = (unsigned char) (t>255 ? 255:t);
  t = l + *(sim+nl+5);
  *(dim+nl+5) = (unsigned char) (t>255 ? 255:t);
  t = l + *(sim+nl+6);
  *(dim+nl+6) = (unsigned char) (t>255 ? 255:t);

  sim += 8; dim += 8; } sim += nl; dim += nl;}}

void resize(void){
  XWindowAttributes xwa;
  XGetWindowAttributes (display, win, &xwa);
  xwa.width  -= xwa.width % 2;
  xwa.height -= xwa.height % 2;
  if(xwa.height != height || xwa.width != width) {
  width  = xwa.width;
  height = xwa.height;
  xim = XCreateImage(display, xwa.visual, xwa.depth, ZPixmap, 0, 0,
		     width, height, 8, 0);
  free(palaka1);
  if(light_on) free(palaka2);
  palaka1 = calloc(xim->height,xim->bytes_per_line);
  palaka2 = palaka1;
  if(light_on) palaka2 = calloc(xim->height,xim->bytes_per_line);
  xim->data = (char *)palaka2;
  if(xim->bits_per_pixel != 32){
  fprintf (stderr,"xim->width = %d \n", xim->width);
  fprintf (stderr,"xim->height = %d \n", xim->height);
  fprintf (stderr,"xim->bytes_per_line = %d \n", xim->bytes_per_line);
  fprintf (stderr,"xim->depth = %d (24)\n", xim->depth);
  fprintf (stderr,"xim->bits_per_pixel = %d (32)\n", xim->bits_per_pixel);
  fprintf (stderr,"xim->xoffset = %d (0)\n", xim->xoffset);
  fprintf (stderr,"xim->bitmap_unit = %d (32)\n", xim->bitmap_unit);
  fprintf (stderr,"xim->red_mask = %.6lx (ff0000)\n", xim->red_mask);
  fprintf (stderr,"xim->green_mask = %.6lx (00ff00)\n", xim->green_mask);
  fprintf (stderr,"xim->blue_mask = %.6lx (0000ff)\n", xim->blue_mask);
  fprintf(stderr,"\nSorry.. This display mode is currently not supported.\n\n");
  exit(0); }}}

void sniff_events(fireshell *fss){
  XEvent e;
  while (XPending(display)){
  XNextEvent (display, &e);
  if (e.type == ConfigureNotify) resize();
  if (e.type == ButtonPress)    {recycle(fss,e.xbutton.x, e.xbutton.y);
                                 memset(palaka1,0,width*height*4);}   
  if (e.type == DestroyNotify)  {fprintf (stderr,"caught DestroyNotify\n"); exit (0);} }}

int main(int argc,char *argv[]){
  unsigned int n,q,screen_num;
  GC gc;
  XGCValues gcv;
  firepix *fpixs, *fpix;
  fireshell *fshells, *fshell;
  for(n=1;n<argc;n++){
  if(strcmp(argv[n],"-nolight")==0) light_on=0; }
  fprintf(stderr,"XFireworks - pyrotechnics simulation program \n");
  fprintf(stderr,"Copyright (c) 1999-2004 Rony B Chandran <ronybc@asia.com> \n\n");
  fprintf(stderr,"For latest version and to drop a small donation, \n");
  fprintf(stderr,"please visit.. http://www.ronybc.8k.com \n");
  display = XOpenDisplay(0); 
  if(!display){ 
  fprintf(stderr,"\nError.. Failed to open display \n\n"); exit(0);}
  screen_num = DefaultScreen (display);

  win = XCreateSimpleWindow(display,
			    RootWindow(display, screen_num),
			    0, 0, WIDTH, HEIGHT, 0, 
                            WhitePixel(display, screen_num),
                            BlackPixel(display, screen_num));

  XSelectInput (display, win, StructureNotifyMask | ButtonPressMask);

  XMapWindow(display, win);
  gc = XCreateGC(display, win, 0, &gcv);

  resize();  /* initialize */

  srandom(time(0));
  
  fpixs = malloc(sizeof(firepix) * PIXCOUNT * SHELLCOUNT);
  fshells = malloc(sizeof(fireshell) * SHELLCOUNT);
  fshell = fshells;
  fpix = fpixs;
  for (n=0;n<SHELLCOUNT;n++){
  fshell->fpix = fpix;
  recycle (fshell,rnd(width),rnd(height));
  fshell++; 
  fpix += PIXCOUNT; }

  while(1) {

  for(q=FTWEAK;q;q--){
  fshell=fshells;
  for(n=SHELLCOUNT;n;n--){
  if (!explode(fshell)){
    recycle(fshell,rnd(width),rnd(height)); }
    fshell++; }}
  
  usleep(15000);   /* for too fast machines */

  if(light_on) light_2x2(fshells);
  XPutImage(display,win,gc,xim,0,0,0,0,width,height);
  XSync(display,0);
  sniff_events(fshells);
  blur(); }

}
